@@ -1,6 +1,7 @@
 #if HAVE_CONFIG_H
 #include <config.h>
 #endif
+#include "data-pool.h"
 #include "maxminddb.h"
 #include "maxminddb-compat-util.h"
 #include <assert.h>
@@ -130,6 +131,9 @@ typedef struct record_info_s {
 /* This is 128kb */
 #define METADATA_BLOCK_MAX_SIZE 131072
 
+// 64 leads us to allocating 4 KiB on a 64bit system.
+#define MMDB_POOL_INIT_SIZE 64
+
 /* *INDENT-OFF* */
 /* --prototypes automatically generated by dev-bin/regen-prototypes.pl - don't remove this comment */
 LOCAL int map_file(MMDB_s *const mmdb);
@@ -176,8 +180,10 @@ LOCAL int decode_one(MMDB_s *mmdb, uint32_t offset,
 LOCAL int get_ext_type(int raw_ext_type);
 LOCAL uint32_t get_ptr_from(uint8_t ctrl, uint8_t const *const ptr,
                             int ptr_size);
-LOCAL int get_entry_data_list(MMDB_s *mmdb, uint32_t offset,
-                              MMDB_entry_data_list_s *const entry_data_list,
+LOCAL int get_entry_data_list(MMDB_s *mmdb,
+                              uint32_t offset,
+                              size_t const index,
+                              MMDB_data_pool_s *const pool,
                               int depth);
 LOCAL float get_ieee754_float(const uint8_t *restrict p);
 LOCAL double get_ieee754_double(const uint8_t *restrict p);
@@ -186,7 +192,6 @@ LOCAL uint32_t get_uint24(const uint8_t *p);
 LOCAL uint32_t get_uint16(const uint8_t *p);
 LOCAL uint64_t get_uintX(const uint8_t *p, int length);
 LOCAL int32_t get_sintX(const uint8_t *p, int length);
-LOCAL MMDB_entry_data_list_s *new_entry_data_list(void);
 LOCAL void free_mmdb_struct(MMDB_s *const mmdb);
 LOCAL void free_languages_metadata(MMDB_s *mmdb);
 LOCAL void free_descriptions_metadata(MMDB_s *mmdb);
@@ -241,7 +246,7 @@ int MMDB_open(const char *const filename, uint32_t flags, MMDB_s *const mmdb)
     }
     mmdb->flags = flags;
 
-    if (MMDB_SUCCESS != (status = map_file(mmdb)) ) {
+    if (MMDB_SUCCESS != (status = map_file(mmdb))) {
         goto cleanup;
     }
 
@@ -297,6 +302,15 @@ int MMDB_open(const char *const filename, uint32_t flags, MMDB_s *const mmdb)
     mmdb->ipv4_start_node.node_value = 0;
     mmdb->ipv4_start_node.netmask = 0;
 
+    // We do this immediately as otherwise there is a race to set
+    // ipv4_start_node.node_value and ipv4_start_node.netmask.
+    if (mmdb->metadata.ip_version == 6) {
+        status = find_ipv4_start_node(mmdb);
+        if (status != MMDB_SUCCESS) {
+            goto cleanup;
+        }
+    }
+
  cleanup:
     if (MMDB_SUCCESS != status) {
         int saved_errno = errno;
@@ -1634,15 +1648,38 @@ int MMDB_get_metadata_as_entry_data_list(
 int MMDB_get_entry_data_list(
     MMDB_entry_s *start, MMDB_entry_data_list_s **const entry_data_list)
 {
-    *entry_data_list = new_entry_data_list();
-    if (NULL == *entry_data_list) {
+    MMDB_data_pool_s *const pool = data_pool_new(MMDB_POOL_INIT_SIZE);
+    if (!pool) {
         return MMDB_OUT_OF_MEMORY_ERROR;
     }
-    return get_entry_data_list(start->mmdb, start->offset, *entry_data_list, 0);
+
+    size_t index = 0;
+    if (data_pool_alloc(pool, &index) != 0) {
+        data_pool_destroy(pool);
+        return MMDB_OUT_OF_MEMORY_ERROR;
+    }
+
+    int const status = get_entry_data_list(start->mmdb, start->offset, index,
+                                           pool, 0);
+
+    *entry_data_list = data_pool_to_list(pool);
+    if (!*entry_data_list) {
+        data_pool_destroy(pool);
+        // This error code is not particularly accurate. However this really
+        // shouldn't happen.
+        return MMDB_OUT_OF_MEMORY_ERROR;
+    }
+
+    // The caller doesn't know about the pool. Just the list.
+    free(pool);
+
+    return status;
 }
 
-LOCAL int get_entry_data_list(MMDB_s *mmdb, uint32_t offset,
-                              MMDB_entry_data_list_s *const entry_data_list,
+LOCAL int get_entry_data_list(MMDB_s *mmdb,
+                              uint32_t offset,
+                              size_t const index,
+                              MMDB_data_pool_s *const pool,
                               int depth)
 {
     if (depth >= MAXIMUM_DATA_STRUCTURE_DEPTH) {
@@ -1650,6 +1687,17 @@ LOCAL int get_entry_data_list(MMDB_s *mmdb, uint32_t offset,
         return MMDB_INVALID_DATA_ERROR;
     }
     depth++;
+
+    // Note it is important to never re-use memory pointing into the pool after
+    // it may have changed in size. This is because the memory location may
+    // move when it grows. In practice this means that if we ever call
+    // data_pool_alloc() or this function recursively, all MMDB_entry_data_list
+    // pointers should be considered invalid, and you should look them up again
+    // by their index.
+    MMDB_entry_data_list_s *entry_data_list = data_pool_lookup(pool, index);
+    if (!entry_data_list) {
+        return MMDB_INVALID_DATA_ERROR;
+    }
     CHECKED_DECODE_ONE(mmdb, offset, &entry_data_list->entry_data);
 
     switch (entry_data_list->entry_data.type) {
@@ -1670,14 +1718,18 @@ LOCAL int get_entry_data_list(MMDB_s *mmdb, uint32_t offset,
             if (entry_data_list->entry_data.type == MMDB_DATA_TYPE_ARRAY
                 || entry_data_list->entry_data.type == MMDB_DATA_TYPE_MAP) {
 
-                int status =
-                    get_entry_data_list(mmdb, last_offset, entry_data_list,
-                                        depth);
+                int status = get_entry_data_list(mmdb, last_offset, index, pool,
+                                                 depth);
                 if (MMDB_SUCCESS != status) {
                     DEBUG_MSG("get_entry_data_list on pointer failed.");
                     return status;
                 }
             }
+
+            entry_data_list = data_pool_lookup(pool, index);
+            if (!entry_data_list) {
+                return MMDB_INVALID_DATA_ERROR;
+            }
             entry_data_list->entry_data.offset_to_next = next_offset;
         }
         break;
@@ -1685,29 +1737,32 @@ LOCAL int get_entry_data_list(MMDB_s *mmdb, uint32_t offset,
         {
             uint32_t array_size = entry_data_list->entry_data.data_size;
             uint32_t array_offset = entry_data_list->entry_data.offset_to_next;
-            MMDB_entry_data_list_s *previous = entry_data_list;
             while (array_size-- > 0) {
-                MMDB_entry_data_list_s *entry_data_list_to = previous->next =
-                                                                 new_entry_data_list();
-                if (NULL == entry_data_list_to) {
+                size_t index2 = 0;
+                if (data_pool_alloc(pool, &index2) != 0) {
                     return MMDB_OUT_OF_MEMORY_ERROR;
                 }
 
-                int status =
-                    get_entry_data_list(mmdb, array_offset, entry_data_list_to,
-                                        depth);
+                int status = get_entry_data_list(mmdb, array_offset, index2,
+                                                 pool, depth);
                 if (MMDB_SUCCESS != status) {
                     DEBUG_MSG("get_entry_data_list on array element failed.");
                     return status;
                 }
 
-                array_offset = entry_data_list_to->entry_data.offset_to_next;
-                while (previous->next) {
-                    previous = previous->next;
+                MMDB_entry_data_list_s *entry_data_list_to
+                    = data_pool_lookup(pool, index2);
+                if (!entry_data_list_to) {
+                    return MMDB_INVALID_DATA_ERROR;
                 }
+                array_offset = entry_data_list_to->entry_data.offset_to_next;
+            }
+
+            entry_data_list = data_pool_lookup(pool, index);
+            if (!entry_data_list) {
+                return MMDB_INVALID_DATA_ERROR;
             }
             entry_data_list->entry_data.offset_to_next = array_offset;
-
         }
         break;
     case MMDB_DATA_TYPE_MAP:
@@ -1715,46 +1770,49 @@ LOCAL int get_entry_data_list(MMDB_s *mmdb, uint32_t offset,
             uint32_t size = entry_data_list->entry_data.data_size;
 
             offset = entry_data_list->entry_data.offset_to_next;
-            MMDB_entry_data_list_s *previous = entry_data_list;
             while (size-- > 0) {
-                MMDB_entry_data_list_s *entry_data_list_to = previous->next =
-                                                                 new_entry_data_list();
-                if (NULL == entry_data_list_to) {
+                size_t key_index = 0;
+                if (data_pool_alloc(pool, &key_index) != 0) {
                     return MMDB_OUT_OF_MEMORY_ERROR;
                 }
 
-                int status =
-                    get_entry_data_list(mmdb, offset, entry_data_list_to,
-                                        depth);
+                int status = get_entry_data_list(mmdb, offset, key_index, pool,
+                                                 depth);
                 if (MMDB_SUCCESS != status) {
                     DEBUG_MSG("get_entry_data_list on map key failed.");
                     return status;
                 }
 
-                while (previous->next) {
-                    previous = previous->next;
+                MMDB_entry_data_list_s *entry_data_list_to
+                    = data_pool_lookup(pool, key_index);
+                if (!entry_data_list_to) {
+                    return MMDB_INVALID_DATA_ERROR;
                 }
-
                 offset = entry_data_list_to->entry_data.offset_to_next;
-                entry_data_list_to = previous->next =
-                                         new_entry_data_list();
 
-                if (NULL == entry_data_list_to) {
+                size_t value_index = 0;
+                if (data_pool_alloc(pool, &value_index) != 0) {
                     return MMDB_OUT_OF_MEMORY_ERROR;
                 }
 
-                status = get_entry_data_list(mmdb, offset, entry_data_list_to,
+                status = get_entry_data_list(mmdb, offset, value_index, pool,
                                              depth);
                 if (MMDB_SUCCESS != status) {
                     DEBUG_MSG("get_entry_data_list on map element failed.");
                     return status;
                 }
 
-                while (previous->next) {
-                    previous = previous->next;
+                entry_data_list_to = data_pool_lookup(pool, value_index);
+                if (!entry_data_list_to) {
+                    return MMDB_INVALID_DATA_ERROR;
                 }
                 offset = entry_data_list_to->entry_data.offset_to_next;
             }
+
+            entry_data_list = data_pool_lookup(pool, index);
+            if (!entry_data_list) {
+                return MMDB_INVALID_DATA_ERROR;
+            }
             entry_data_list->entry_data.offset_to_next = offset;
         }
         break;
@@ -1832,21 +1890,11 @@ LOCAL int32_t get_sintX(const uint8_t *p, int length)
     return (int32_t)get_uintX(p, length);
 }
 
-LOCAL MMDB_entry_data_list_s *new_entry_data_list(void)
-{
-    /* We need calloc here in order to ensure that the ->next pointer in the
-     * struct doesn't point to some random address. */
-    return calloc(1, sizeof(MMDB_entry_data_list_s));
-}
-
 void MMDB_free_entry_data_list(MMDB_entry_data_list_s *const entry_data_list)
 {
     if (entry_data_list == NULL) {
         return;
     }
-    if (entry_data_list->next) {
-        MMDB_free_entry_data_list(entry_data_list->next);
-    }
     free(entry_data_list);
 }
 
